@@ -9,30 +9,27 @@ from sql import Null
 from sql.aggregate import Sum
 from sql.conditionals import Coalesce
 
-from trytond.model import ModelView, Workflow, ModelSQL, fields, Check, Unique
+from trytond.i18n import gettext
+from trytond.model import ModelView, Workflow, ModelSQL, fields, Unique
+from trytond.model.exceptions import AccessError
 from trytond.wizard import Wizard, StateView, StateTransition, Button
 from trytond.pyson import Not, Equal, Eval, Or, Bool, If
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 from trytond.tools import reduce_ids, grouped_slice
 
-__all__ = ['Forecast', 'ForecastLine', 'ForecastLineMove',
-    'ForecastCompleteAsk', 'ForecastCompleteChoose', 'ForecastComplete']
-
-STATES = {
-    'readonly': Not(Equal(Eval('state'), 'draft')),
-}
-DEPENDS = ['state']
-FORECAST_STATES = [
-    ('draft', 'Draft'),
-    ('done', 'Done'),
-    ('cancel', 'Cancel'),
-    ]
+from .exceptions import ForecastValidationError
 
 
 class Forecast(Workflow, ModelSQL, ModelView):
     "Stock Forecast"
     __name__ = "stock.forecast"
+
+    _states = {
+        'readonly': Not(Equal(Eval('state'), 'draft')),
+    }
+    _depends = ['state']
+
     warehouse = fields.Many2One(
         'stock.location', 'Location', required=True,
         domain=[('type', '=', 'warehouse')], states={
@@ -42,47 +39,45 @@ class Forecast(Workflow, ModelSQL, ModelView):
         depends=['state'])
     destination = fields.Many2One(
         'stock.location', 'Destination', required=True,
-        domain=[('type', 'in', ['customer', 'production'])], states=STATES,
-        depends=DEPENDS)
-    from_date = fields.Date('From Date', required=True, states=STATES,
-        depends=DEPENDS)
-    to_date = fields.Date('To Date', required=True, states=STATES,
-        depends=DEPENDS)
+        domain=[('type', 'in', ['customer', 'production'])], states=_states,
+        depends=_depends)
+    from_date = fields.Date(
+        "From Date", required=True,
+        domain=[('from_date', '<=', Eval('to_date'))],
+        states=_states, depends=_depends + ['to_date'])
+    to_date = fields.Date(
+        "To Date", required=True,
+        domain=[('to_date', '>=', Eval('from_date'))],
+        states=_states, depends=_depends + ['from_date'])
     lines = fields.One2Many(
-        'stock.forecast.line', 'forecast', 'Lines', states=STATES,
-        depends=DEPENDS)
+        'stock.forecast.line', 'forecast', 'Lines', states=_states,
+        depends=_depends)
     company = fields.Many2One(
         'company.company', 'Company', required=True, states={
             'readonly': Or(Not(Equal(Eval('state'), 'draft')),
                 Bool(Eval('lines', [0]))),
             },
         depends=['state'])
-    state = fields.Selection(
-        FORECAST_STATES, 'State', readonly=True, select=True)
+    state = fields.Selection([
+            ('draft', "Draft"),
+            ('done', "Done"),
+            ('cancelled', "Cancelled"),
+            ], "State", readonly=True, select=True)
     active = fields.Function(fields.Boolean('Active'),
         'get_active', searcher='search_active')
 
+    del _states, _depends
+
     @classmethod
     def __setup__(cls):
         super(Forecast, cls).__setup__()
-        t = cls.__table__()
-        cls._sql_constraints += [
-            ('check_from_to_date', Check(t, t.to_date >= t.from_date),
-                '"To Date" must be greater than "From Date"'),
-            ]
-        cls._error_messages.update({
-                'date_overlap': ('Forecast "%(first)s" overlaps with dates '
-                    'of forecast "%(second)s" in the same location.'),
-                'delete_cancel': ('Forecast "%s" must be cancelled before '
-                    'deletion.'),
-                })
         cls._order.insert(0, ('from_date', 'DESC'))
         cls._order.insert(1, ('warehouse', 'ASC'))
         cls._transitions |= set((
                 ('draft', 'done'),
-                ('draft', 'cancel'),
+                ('draft', 'cancelled'),
                 ('done', 'draft'),
-                ('cancel', 'draft'),
+                ('cancelled', 'draft'),
                 ))
         cls._buttons.update({
                 'cancel': {
@@ -106,12 +101,23 @@ class Forecast(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        cursor = Transaction().connection.cursor()
+        sql_table = cls.__table__()
+
         super(Forecast, cls).__register__(module_name)
 
         # Add index on create_date
         table = cls.__table_handler__(module_name)
         table.index_action('create_date', action='add')
 
+        # Migration from 5.0: remove check_from_to_date
+        table.drop_constraint('check_from_to_date')
+
+        # Migration from 5.6: rename state cancel to cancelled
+        cursor.execute(*sql_table.update(
+                [sql_table.state], ['cancelled'],
+                where=sql_table.state == 'cancel'))
+
     @staticmethod
     def default_state():
         return 'draft'
@@ -189,18 +195,20 @@ class Forecast(Workflow, ModelSQL, ModelView):
         forecast_id = cursor.fetchone()
         if forecast_id:
             second = self.__class__(forecast_id[0])
-            self.raise_user_error('date_overlap', {
-                    'first': self.rec_name,
-                    'second': second.rec_name,
-                    })
+            raise ForecastValidationError(
+                gettext('stock_forecast.msg_forecast_date_overlap',
+                    first=self.rec_name,
+                    second=second.rec_name))
 
     @classmethod
     def delete(cls, forecasts):
         # Cancel before delete
         cls.cancel(forecasts)
         for forecast in forecasts:
-            if forecast.state != 'cancel':
-                cls.raise_user_error('delete_cancel', forecast.rec_name)
+            if forecast.state != 'cancelled':
+                raise AccessError(
+                    gettext('stock_forecast.msg_forecast_delete_cancel',
+                        forecast=forecast.rec_name))
         super(Forecast, cls).delete(forecasts)
 
     @classmethod
@@ -217,7 +225,7 @@ class Forecast(Workflow, ModelSQL, ModelView):
 
     @classmethod
     @ModelView.button
-    @Workflow.transition('cancel')
+    @Workflow.transition('cancelled')
     def cancel(cls, forecasts):
         pass
 
@@ -273,11 +281,14 @@ class ForecastLine(ModelSQL, ModelView):
         depends=['product', 'product_uom_category'] + _depends)
     unit_digits = fields.Function(fields.Integer('Unit Digits'),
             'get_unit_digits')
-    quantity = fields.Float('Quantity', digits=(16, Eval('unit_digits', 2)),
-        required=True, states=_states, depends=['unit_digits'] + _depends)
-    minimal_quantity = fields.Float('Minimal Qty',
-        digits=(16, Eval('unit_digits', 2)), required=True,
+    quantity = fields.Float(
+        "Quantity", digits=(16, Eval('unit_digits', 2)), required=True,
+        domain=[('quantity', '>=', 0)],
         states=_states, depends=['unit_digits'] + _depends)
+    minimal_quantity = fields.Float(
+        "Minimal Qty", digits=(16, Eval('unit_digits', 2)), required=True,
+        domain=[('minimal_quantity', '<=', Eval('quantity'))],
+        states=_states, depends=['unit_digits', 'quantity'] + _depends)
     moves = fields.Many2Many('stock.forecast.line-stock.move',
         'line', 'move', 'Moves', readonly=True)
     forecast = fields.Many2One(
@@ -288,7 +299,7 @@ class ForecastLine(ModelSQL, ModelView):
             },
         depends=['forecast_state'])
     forecast_state = fields.Function(
-        fields.Selection(FORECAST_STATES, 'Forecast State'),
+        fields.Selection('get_forecast_states', 'Forecast State'),
         'on_change_with_forecast_state')
     quantity_executed = fields.Function(fields.Float('Quantity Executed',
             digits=(16, Eval('unit_digits', 2)), depends=['unit_digits']),
@@ -299,17 +310,23 @@ class ForecastLine(ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(ForecastLine, cls).__setup__()
+        cls.__access__.add('forecast')
         t = cls.__table__()
         cls._sql_constraints += [
-            ('check_line_qty_pos', Check(t, t.quantity >= 0),
-                'Line quantity must be positive'),
-            ('check_line_minimal_qty',
-                Check(t, t.quantity >= t.minimal_quantity),
-                'Line quantity must be greater than the minimal quantity'),
             ('forecast_product_uniq', Unique(t, t.forecast, t.product),
-                'Product must be unique by forecast'),
+                'stock_forecast.msg_forecast_line_product_unique'),
             ]
 
+    @classmethod
+    def __register__(cls, module_name):
+        super().__register__(module_name)
+
+        table_h = cls.__table_handler__(module_name)
+
+        # Migration from 5.0: remove check on quantity
+        table_h.drop_constraint('check_line_qty_pos')
+        table_h.drop_constraint('check_line_minimal_qty')
+
     @staticmethod
     def default_unit_digits():
         return 2
@@ -339,6 +356,12 @@ class ForecastLine(ModelSQL, ModelView):
     def get_unit_digits(self, name):
         return self.product.default_uom.digits
 
+    @classmethod
+    def get_forecast_states(cls):
+        pool = Pool()
+        Forecast = pool.get('stock.forecast')
+        return Forecast.fields_get(['state'])['state']['selection']
+
     @fields.depends('forecast', '_parent_forecast.state')
     def on_change_with_forecast_state(self, name=None):
         if self.forecast:
@@ -367,7 +390,9 @@ class ForecastLine(ModelSQL, ModelView):
         line_move = LineMove.__table__()
 
         result = dict((x.id, 0) for x in lines)
-        key = lambda line: line.forecast.id
+
+        def key(line):
+            return line.forecast.id
         lines.sort(key=key)
         for forecast_id, lines in itertools.groupby(lines, key):
             forecast = Forecast(forecast_id)
@@ -387,14 +412,14 @@ class ForecastLine(ModelSQL, ModelView):
                         & (location_from.right <= forecast.warehouse.right)
                         & (location_to.left >= forecast.destination.left)
                         & (location_to.right <= forecast.destination.right)
-                        & (move.state != 'cancel')
+                        & (move.state != 'cancelled')
                         & (Coalesce(move.effective_date, move.planned_date)
                             >= forecast.from_date)
                         & (Coalesce(move.effective_date, move.planned_date)
                             <= forecast.to_date)
                         & (line_move.id == Null),
                         group_by=move.product))
-                for product_id, quantity in cursor.fetchall():
+                for product_id, quantity in cursor:
                     line = product2line[product_id]
                     result[line.id] = Uom.compute_qty(line.product.default_uom,
                         quantity, line.uom)
@@ -501,8 +526,14 @@ class ForecastLineMove(ModelSQL):
 class ForecastCompleteAsk(ModelView):
     'Complete Forecast'
     __name__ = 'stock.forecast.complete.ask'
-    from_date = fields.Date('From Date', required=True)
-    to_date = fields.Date('To Date', required=True)
+    from_date = fields.Date(
+        "From Date", required=True,
+        domain=[('from_date', '<', Eval('to_date'))],
+        depends=['to_date'])
+    to_date = fields.Date(
+        "To Date", required=True,
+        domain=[('to_date', '>', Eval('from_date'))],
+        depends=['from_date'])
 
 
 class ForecastCompleteChoose(ModelView):
@@ -529,39 +560,25 @@ class ForecastComplete(Wizard):
             ])
     complete = StateTransition()
 
-    @classmethod
-    def __setup__(cls):
-        super(ForecastComplete, cls).__setup__()
-        cls._error_messages.update({
-                'from_to_date': (
-                    '"From Date" should be smaller than "To Date".'),
-                })
-
     def default_ask(self, fields):
         """
         Forecast dates shifted by one year.
         """
-        Forecast = Pool().get('stock.forecast')
-        forecast = Forecast(Transaction().context['active_id'])
-
-        res = {}
+        default = {}
         for field in ("to_date", "from_date"):
-            res[field] = getattr(forecast, field) - relativedelta(years=1)
-        return res
+            default[field] = (
+                getattr(self.record, field) - relativedelta(years=1))
+        return default
 
     def _get_product_quantity(self):
         pool = Pool()
-        Forecast = pool.get('stock.forecast')
         Product = pool.get('product.product')
-        forecast = Forecast(Transaction().context['active_id'])
-        if self.ask.from_date > self.ask.to_date:
-            self.raise_user_error('from_to_date')
 
         with Transaction().set_context(
-                stock_destinations=[forecast.destination.id],
+                stock_destinations=[self.record.destination.id],
                 stock_date_start=self.ask.from_date,
                 stock_date_end=self.ask.to_date):
-            return Product.products_by_location([forecast.warehouse.id],
+            return Product.products_by_location([self.record.warehouse.id],
                 with_childs=True)
 
     def default_choose(self, fields):
@@ -580,11 +597,10 @@ class ForecastComplete(Wizard):
 
     def transition_complete(self):
         pool = Pool()
-        Forecast = pool.get('stock.forecast')
         ForecastLine = pool.get('stock.forecast.line')
         Product = pool.get('product.product')
 
-        forecast = Forecast(Transaction().context['active_id'])
+        forecast = self.record
         prod2line = {}
         forecast_lines = ForecastLine.search([
                 ('forecast', '=', forecast.id),
